@@ -27,20 +27,20 @@ _pkg_config = {
     'default_sip_dir':  '/opt/kde/share/sip/',
     'py_inc_dir':       '/usr/include/python2.4',
     'py_lib_dir':       '/usr/lib/python2.4/config',
-    'py_version':       0x020400,
+    'py_version':       0x020401,
     'qt_dir':           '/usr/lib/qt/',
     'qt_edition':       'free',
-    'qt_inc_dir':       '/usr/lib/qt//include',
+    'qt_inc_dir':       '/usr/lib/qt/include',
     'qt_lib':           'qt-mt',
     'qt_lib_dir':       '/usr/lib/qt//lib',
     'qt_threaded':      1,
-    'qt_version':       0x030303,
+    'qt_version':       0x030304,
     'qt_winconfig':     '',
     'sip_bin':          '/opt/kde/bin/sip',
     'sip_inc_dir':      '/usr/include/python2.4',
     'sip_mod_dir':      '/usr/lib/python2.4/site-packages',
-    'sip_version':      0x040001,
-    'sip_version_str':  '4.0.1'
+    'sip_version':      0x040101,
+    'sip_version_str':  '4.1.1'
 }
 
 _default_macros = {
@@ -91,7 +91,7 @@ _default_macros = {
     'EXTENSION_SHLIB':          '',
     'INCDIR':                   '',
     'INCDIR_OPENGL':            '/usr/X11R6/include',
-    'INCDIR_QT':                '/usr/lib/qt//include',
+    'INCDIR_QT':                '/usr/lib/qt/include',
     'INCDIR_X11':               '/usr/X11R6/include',
     'LFLAGS':                   '',
     'LFLAGS_CONSOLE':           '',
@@ -257,7 +257,7 @@ class Makefile:
     """
     def __init__(self, configuration, console=0, qt=0, opengl=0, python=0,
                  threaded=0, warnings=None, debug=0, dir=None,
-                 makefile="Makefile", installs=None):
+                 makefile="Makefile", installs=None, srcdir=None, builddir=None):
         """Initialise an instance of the target.  All the macros are left
         unchanged allowing scripts to manipulate them at will.
 
@@ -293,6 +293,13 @@ class Makefile:
         self._dir = dir
         self._makefile = makefile
         self._installs = installs
+        
+        if srcdir is None:
+            srcdir = os.getcwd()
+        self._srcdir = srcdir
+        if builddir is None:
+            builddir = os.getcwd()
+        self._builddir = builddir
 
         self._finalised = 0
 
@@ -391,7 +398,11 @@ class Makefile:
         defines.extend(self.extra_defines)
         defines.extend(self.optional_list("DEFINES"))
 
-        incdir = _UniqueList(["."])
+        if self._dir is None:
+            srcdir = self._srcdir
+        else:
+            srcdir = os.path.join(self._srcdir,self._dir)
+        incdir = _UniqueList([srcdir])
         incdir.extend(self.extra_include_dirs)
         incdir.extend(self.optional_list("INCDIR"))
 
@@ -444,7 +455,7 @@ class Makefile:
 
                 libs.append(self.platform_lib(py_lib))
 
-        if self.generator in ("MSVC", "BMAKE"):
+        if self.generator in ("MSVC", "MSVC.NET", "BMAKE"):
             if win_exceptions:
                 cflags_exceptions = "CFLAGS_EXCEPTIONS_ON"
                 cxxflags_exceptions = "CXXFLAGS_EXCEPTIONS_ON"
@@ -498,7 +509,7 @@ class Makefile:
             cxxflags_debug = "CXXFLAGS_RELEASE"
             lflags_debug = "LFLAGS_RELEASE"
 
-        if self.generator in ("MSVC", "BMAKE"):
+        if self.generator in ("MSVC", "MSVC.NET", "BMAKE"):
             if self._threaded:
                 cflags.extend(self.optional_list(cflags_mt))
                 cxxflags.extend(self.optional_list(cxxflags_mt))
@@ -545,7 +556,7 @@ class Makefile:
             # Windows needs the version number appended if Qt is a DLL.
             qt_lib = self.config.qt_lib
 
-            if self.generator in ("MSVC", "BMAKE") and win_shared:
+            if self.generator in ("MSVC", "MSVC.NET", "BMAKE") and win_shared:
                 qt_lib = qt_lib + string.replace(version_to_string(self.config.qt_version), ".", "")
 
                 if self.config.qt_edition == "non-commercial":
@@ -643,7 +654,7 @@ class Makefile:
 
         clib is the library name in cannonical form.
         """
-        if self.generator in ("MSVC", "BMAKE"):
+        if self.generator in ("MSVC", "MSVC.NET", "BMAKE"):
             plib = clib + ".lib"
         else:
             plib = "-l" + clib
@@ -658,7 +669,7 @@ class Makefile:
         """
         prl_libs = []
 
-        if self.generator in ("MSVC", "BMAKE"):
+        if self.generator in ("MSVC", "MSVC.NET", "BMAKE"):
             prl_name = os.path.join(self.config.qt_lib_dir, clib + ".prl")
         else:
             prl_name = os.path.join(self.config.qt_lib_dir, "lib" + clib + ".prl")
@@ -737,13 +748,12 @@ class Makefile:
                 dict[i] = ""
 
         # Generate the list of objects.
-        if self.generator in ("MSVC", "BMAKE"):
+        if self.generator in ("MSVC", "MSVC.NET", "BMAKE"):
             ext = ".obj"
         else:
             ext = ".o"
 
         olist = []
-
         for f in string.split(dict["sources"]):
             root, discard = os.path.splitext(f)
             olist.append(root + ext)
@@ -754,9 +764,13 @@ class Makefile:
 
             root, discard = os.path.splitext(f)
             olist.append("moc_" + root + ext)
-
         dict["objects"] = string.join(olist)
 
+        headers = []
+        for h in string.split(dict["headers"]):
+            headers.append(os.path.join("$(srcdir)",h))
+        dict["headers"] = string.join(headers)
+
         return dict
 
     def clean_build_file_objects(self, mfile, build):
@@ -786,19 +800,25 @@ class Makefile:
         self.ready()
 
         if self._dir:
-            mfname = os.path.join(self._dir, self._makefile)
+            mfname = os.path.join(self._builddir, self._dir, self._makefile)
+            builddir = os.path.join(self._builddir, self._dir)
+            if not os.path.exists(builddir):
+                os.mkdir(builddir)
+            srcdir = os.path.join(self._srcdir,self._dir)
         else:
-            mfname = self._makefile
-
+            mfname = os.path.join(self._builddir, self._makefile)
+            srcdir = self._srcdir
+            
         try:
             mfile = open(mfname, "w")
         except IOError, detail:
             error("Unable to create \"%s\": %s" % (mfname, detail))
 
+        mfile.write("srcdir = %s\n" % srcdir)
         self.generate_macros_and_rules(mfile)
         self.generate_target_default(mfile)
-
         self.generate_target_install(mfile)
+
         if self._installs:
             if type(self._installs) != types.ListType:
                 self._installs = [self._installs]
@@ -844,7 +864,7 @@ class Makefile:
 
         libs = []
 
-        if self.generator == "MSVC":
+        if self.generator in ("MSVC", "MSVC.NET"):
             libdir_prefix = "/LIBPATH:"
         else:
             libdir_prefix = "-L"
@@ -867,7 +887,8 @@ class Makefile:
         else:
             mfile.write(".SUFFIXES: .c .cpp .cc .cxx .C\n\n")
 
-        if self.generator == "MSVC":
+        if self.generator in ("MSVC", "MSVC.NET"):
+            # FIXME srcdir!=builddir support
             mfile.write("""
 {.}.cpp{}.obj::
 \t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -Fo @<<
@@ -895,6 +916,7 @@ class Makefile:
 <<
 """)
         elif self.generator == "BMAKE":
+            # FIXME srcdir!=builddir support
             mfile.write("""
 .cpp.obj:
 \t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -o$@ $<
@@ -913,19 +935,19 @@ class Makefile:
 """)
         else:
             mfile.write("""
-.cpp.o:
+%.o : $(srcdir)/%.cpp
 \t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -o $@ $<
 
-.cc.o:
+%.o : $(srcdir)/%.cc
 \t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -o $@ $<
 
-.cxx.o:
+%.o : $(srcdir)/%.cxx
 \t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -o $@ $<
 
-.C.o:
+%.o : $(srcdir)/%.C
 \t$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) -o $@ $<
 
-.c.o:
+%.o : $(srcdir)/%.c
 \t$(CC) -c $(CFLAGS) $(CPPFLAGS) -o $@ $<
 """)
 
@@ -1011,12 +1033,12 @@ class ParentMakefile(Makefile):
     """The class that represents a parent Makefile.
     """
     def __init__(self, configuration, subdirs, dir=None, makefile="Makefile",
-                 installs=None):
+                 installs=None, srcdir=None, builddir=None):
         """Initialise an instance of a parent Makefile.
 
         subdirs is the sequence of subdirectories.
         """
-        Makefile.__init__(self, configuration, dir=dir, makefile=makefile, installs=installs)
+        Makefile.__init__(self, configuration, dir=dir, makefile=makefile, installs=installs, srcdir=srcdir, builddir=builddir)
 
         self._subdirs = subdirs
 
@@ -1118,14 +1140,15 @@ class ModuleMakefile(Makefile):
     """
     def __init__(self, configuration, build_file, install_dir=None, static=0,
                  console=0, qt=0, opengl=0, threaded=0, warnings=None, debug=0,
-                 dir=None, makefile="Makefile", installs=None):
+                 dir=None, makefile="Makefile", installs=None, srcdir=None,
+                 builddir=None):
         """Initialise an instance of a module Makefile.
 
         build_file is the file containing the target specific information.
         install_dir is the directory the target will be installed in.
         static is set if the module should be built as a static library.
         """
-        Makefile.__init__(self, configuration, console, qt, opengl, 1, threaded, warnings, debug, dir, makefile, installs)
+        Makefile.__init__(self, configuration, console, qt, opengl, 1, threaded, warnings, debug, dir, makefile, installs, srcdir, builddir)
 
         self._build = self.parse_build_file(build_file)
         self._install_dir = install_dir
@@ -1248,7 +1271,7 @@ class ModuleMakefile(Makefile):
             if sys.platform != "win32":
                 mname = mname + "module"
 
-        if self.generator in ("MSVC", "BMAKE"):
+        if self.generator in ("MSVC", "MSVC.NET", "BMAKE"):
             mname = "lib" + mname
 
         return mname
@@ -1277,7 +1300,7 @@ class ModuleMakefile(Makefile):
         mfile.write("\n")
 
         if self.static:
-            if self.generator in ("MSVC", "BMAKE"):
+            if self.generator in ("MSVC", "MSVC.NET", "BMAKE"):
                 mfile.write("LIB = %s\n" % self.required_string("LIB"))
             else:
                 mfile.write("AR = %s\n" % self.required_string("AR"))
@@ -1296,7 +1319,7 @@ class ModuleMakefile(Makefile):
         """
         mfile.write("\n$(TARGET): $(OFILES)\n")
 
-        if self.generator == "MSVC":
+        if self.generator in ("MSVC", "MSVC.NET"):
             if self.static:
                 mfile.write("\t$(LIB) /OUT:$(TARGET) @<<\n")
                 mfile.write("\t  $(OFILES)\n")
@@ -1368,7 +1391,7 @@ class ModuleMakefile(Makefile):
         if self.config.sip_version < 0x040000 and not self.static:
             if self.generator == "MINGW":
                 self.install_file(mfile, self._target + ".a", self._install_dir)
-            elif self.generator == "MSVC":
+            elif self.generator in ("MSVC", "MSVC.NET"):
                 self.install_file(mfile, self._target + ".lib", self._install_dir)
                 self.install_file(mfile, self._target + ".exp", self._install_dir)
             elif self.generator == "BMAKE":
@@ -1434,13 +1457,13 @@ class ProgramMakefile(Makefile):
     def __init__(self, configuration, build_file=None, install_dir=None,
                  console=0, qt=0, opengl=0, python=0, threaded=0,
                  warnings=None, debug=0, dir=None, makefile="Makefile",
-                 installs=None):
+                 installs=None, srcdir=None, builddir=None):
         """Initialise an instance of a program Makefile.
 
         build_file is the file containing the target specific information.
         install_dir is the directory the target will be installed in.
         """
-        Makefile.__init__(self, configuration, console, qt, opengl, python, threaded, warnings, debug, dir, makefile, installs)
+        Makefile.__init__(self, configuration, console, qt, opengl, python, threaded, warnings, debug, dir, makefile, installs, srcdir, builddir)
 
         self._install_dir = install_dir
 
@@ -1462,7 +1485,8 @@ class ProgramMakefile(Makefile):
 
         if sys.platform == "win32":
             exe = exe + ".exe"
-
+        exe = os.path.join(self._srcdir,exe)
+        
         # The command line.
         build = []
 
@@ -1478,9 +1502,9 @@ class ProgramMakefile(Makefile):
 
         # Borland requires all flags to precede all file names.
         if self.generator != "BMAKE":
-            build.append(source)
+            build.append(os.path.join(self._srcdir,source))
 
-        if self.generator == "MSVC":
+        if self.generator in ("MSVC", "MSVC.NET"):
             build.append("-Fe")
             build.append("/link")
             libdir_prefix = "/LIBPATH:"
@@ -1523,14 +1547,14 @@ class ProgramMakefile(Makefile):
         build.extend(self.optional_list("LIBS"))
 
         if self.generator == "BMAKE":
-            build.append(source)
+            build.append(os.path.join(source))
 
         return (exe, string.join(build))
 
     def finalise(self):
         """Finalise the macros for a program Makefile.
         """
-        if self.generator == "MSVC":
+        if self.generator in ("MSVC", "MSVC.NET"):
             self.LFLAGS.append("/INCREMENTAL:NO")
 
         if self.console:
@@ -1569,7 +1593,7 @@ class ProgramMakefile(Makefile):
         """
         mfile.write("\n$(TARGET): $(OFILES)\n")
 
-        if self.generator == "MSVC":
+        if self.generator in ("MSVC", "MSVC.NET"):
             mfile.write("\t$(LINK) $(LFLAGS) /OUT:$(TARGET) @<<\n")
             mfile.write("\t  $(OFILES) $(LIBS)\n")
             mfile.write("<<\n")
@@ -1672,7 +1696,8 @@ def read_version(filename, description, numdefine=None, strdefine=None):
                 need_num = 0
 
             if need_str and wl[1] == strdefine:
-                versstr = wl[2][1:-1]
+                # Take account of embedded spaces.
+                versstr = string.join(wl[2:])[1:-1]
                 need_str = 0
 
         l = f.readline()
